/*
* srlua.c
* Lua interpreter for self-running programs
* Luiz Henrique de Figueiredo <lhf@tecgraf.puc-rio.br>
* 26 Aug 2004 09:11:03
* This code is hereby placed in the public domain.
*/

#ifdef _WIN32
#include <windows.h>
#endif

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "glue.h"
#include "lua.h"
#include "lualib.h"
#include "lauxlib.h"

/* name of global that holds table with builtin packages */
#define BUILTTAB	"_BUILTIN"

static void openlibs(lua_State *L)
{
 int t=lua_gettop(L);
 luaopen_base(L);
 luaopen_table(L);
 luaopen_io(L);
 luaopen_string(L);
 luaopen_math(L);
 luaopen_debug(L);
 luaopen_loadlib(L);
 /* add your libraries here */
 luaopen_bit(L);
 luaopen_pack(L);
 luaopen_lfs(L);

 lua_settop(L,t);
}

typedef struct
{
 FILE *f;
 long size;
 char buff[512];
} State;

static const char *myget(lua_State *L, void *data, size_t *size)
{
 State* s=data;
 size_t n;
 if (s->size<=0) return NULL;
 n=(sizeof(s->buff)<=s->size)? sizeof(s->buff) : s->size;
 n=fread(s->buff,1,n,s->f);
 if (n==-1) return NULL;
 s->size-=n;
 *size=n;
 return s->buff;
}

#define cannot(x) luaL_error(L,"cannot %s %s: %s",x,name,strerror(errno))

static void dosegment(lua_State *L, State *pS, const char *name)
{
   if (lua_load(L, myget, pS, name) != 0) lua_error(L);

   lua_getglobal(L, BUILTTAB);
   if (!lua_istable(L, -1)) luaL_error(L, "`" BUILTTAB "' is not a table");
   lua_pushstring(L, name);
   lua_rawget(L, -2);
   /* is it there? */
   if (!lua_toboolean(L, -1)) {
      /* not found, add it */
      lua_pop(L, 1); /* pop nil value */
      lua_pushstring(L, name);
      lua_pushboolean(L, 1);
      lua_rawset(L, -3); /* which is what rawset wants */
   } else {
      /* pop of result of lookup */
      lua_pop(L, 1);
   }
   /* pop table */
   lua_pop(L, 1);
   /* execute chunk which should now be TOS */
   lua_call(L,0,0);
}

static void load(lua_State *L, const char *name)
{
   struct {
      Glue glue;
      const char sig[GLUELEN];
   } t;
   Glue *e;
   State S;
   int i,gsize;
   FILE *f=fopen(name,"rb");
   if (f==NULL) cannot("open");
   if (fseek(f,-sizeof(t),SEEK_END)!=0) cannot("seek");
   if (fread(&t,sizeof(t),1,f)!=1) cannot("read");
   if (memcmp(t.sig,GLUESIG,GLUELEN)!=0) luaL_error(L,"no Lua program found in %s",name);
   
   /* load glue info */
   if (ftell(f) == -1) cannot("tell");
   gsize = ftell(f) - t.glue.start;
   e = malloc(gsize);
   if (fseek(f, t.glue.start, SEEK_SET) != 0) cannot("seek");
   if (fread(e, gsize, 1, f) != 1) cannot("read");
   
   /* handle each segment */
   for (i = 0; i < t.glue.name; i++) {
      if (fseek(f, e[i].start, SEEK_SET) != 0) cannot("seek");
      S.f = f;
      if (i + 1 == t.glue.name)
         /* last segment start index is in tail */
         S.size = t.glue.start - e[i].start;
      else 
         /* otherwise it's in the next segment start */
         S.size = e[i+1].start - e[i].start;
   
      /* load and execute chunk */
      dosegment(L, &S, (char *)e + e[i].name);
   }
   free(e);
}

static int pmain(lua_State *L)
{
 char **argv=lua_touserdata(L,1);
 int i;
 openlibs(L);
 lua_pushliteral(L,"arg");
 lua_newtable(L);
 for (i=0; argv[i]; i++)
 {
  lua_pushstring(L,argv[i]);
  lua_rawseti(L,-2,i);
 }
 lua_pushliteral(L,"n");
 lua_pushnumber(L,i-1);
 lua_rawset(L,-3);
 lua_rawset(L,LUA_GLOBALSINDEX);
 lua_settop(L, 1);
 load(L,argv[0]);
 return 0;
}

#define report(s) fprintf(stderr,"%s: %s\n",argv[0],s)

int main(int argc, char *argv[])
{
 lua_State *L;
#ifdef _WIN32
 char name[MAX_PATH];
 if (GetModuleFileName(NULL,name,sizeof(name))==0)
 {
  report("cannot locate this executable");
  return EXIT_FAILURE;
 }
 argv[0]=name;
#endif
 L=lua_open();
 if (L==NULL)
 {
  report("not enough memory for state");
  return EXIT_FAILURE;
 }
 if (lua_cpcall(L,pmain,argv)) report(lua_tostring(L,-1));
 lua_close(L);
 return EXIT_SUCCESS;
}
